/*
* Copyright 2026 Axel Waggershauser
*/
// SPDX-License-Identifier: Apache-2.0

#pragma once

#include "Version.h"

#if defined(_MSC_VER)
#pragma message("<ZXing/ZXingVersion.h> is deprecated, include ZXing/Version.h instead")
#else
#warning "<ZXing/ZXingVersion.h> is deprecated, include ZXing/Version.h instead"
#endif
