/*
* Copyright 2024 Axel Waggershauser
*/
// SPDX-License-Identifier: Apache-2.0

#pragma once

#include "BarcodeFormat.h"
#include "CreateBarcode.h"
#include "ReadBarcode.h"
#include "WriteBarcode.h"
#include "Version.h"

namespace ZXing {

const std::string& Version();

} // namespace ZXing
